package sgfThumbnailer;
use strict;
use warnings;
use Carp;
use List::Util qw(max min);
use SimpleSGF qw(process_sgf);
use ThumbIM;
use ThumbGD;

sub new {
  my ($this, $module) = @_;
  my $class = ref($this) || $this;
  my $self = {};
  $self->{maxmoves} = 25;
  $self->{scale} = 128;
  $self->{outline} = 1;
  $self->{stonesize} = 20;
  $self->{extension} = 0.12;
  $self->{module} = '';
  $self->{checkselfcatpure} = 0;
  $self->{stonefactor} = 0.9;
  $self->{afontkeys} = [qw(front back imfactor gdfactor xnudge ynudge under black white)];
  $self->{colourhash} = {tsumego => 'cfd5e5',nomove => 'ff739b',game => 'c9c5a4',line => '000000',bs => '000000',ws => 'ffffff',bso => '000000',wso => '000000', bunder => 'ffffff', wunder => 'ffffff', tsumegooff => 'c7f4f3', nomoveoff => 'ffffff', gameoff => 'f39d2a', offboard => 'ffffff'};
  initlib($self, $module);
  bless $self, $class;
}

sub initlib {
  my ($self, $module) = @_;
  if (defined $module) {
    setlib($self, $module);
  } else {
    my $lib = 'GD';
    eval "require $lib";
    if ($@) {
      $lib = 'Image::Magick';
      eval "require $lib";
      croak 'No supported graphics library found' if $@;
      $self->{code} = new ThumbIM;
    } else {
      $self->{code} = new ThumbGD;
    }
    $self->{module} = $lib;
  }
}

sub setlib {
  my ($self, $module) = @_;
  if (defined $module) {
    if ($module eq 'Image::Magick' || $module eq 'GD') {
      if ($module ne $self->{module}) {
        eval "require $module";
        croak $@ if $@;
        $self->{module} = $module;
        if ($module eq 'GD') {
          $self->{code} = new ThumbGD;
        } else {
          $self->{code} = new ThumbIM;
        }
      }
    } else {
      croak "Unsupported graphics library $module";
    }
  }
  return $self->{module};
}

sub realboard {
  my ($self, $rflag) = @_;
  return set($self, 'rflag', $rflag);
}

sub extension {
  my ($self, $extension) = @_;
  if (!isnumeric($extension) || $extension < 0) {
    croak "Extension must be a positive number (got $extension)";
  }
  return set($self, 'extension', $extension);
}

sub outline {
  my ($self, $outline) = @_;
  return set($self, 'outline', $outline);
}

sub stonesize {
  my ($self, $stonefactor) = @_;
  if (!isnumeric($stonefactor) || $stonefactor < 0) {
    croak "Stonesize must be a positive number (got $stonefactor)";
  }
  return set($self, 'stonefactor', $stonefactor);
}

sub block {
  my ($self, $blockfontpath, $hashref) = @_;
  if (-e $blockfontpath) {
    $self->{block} = $hashref;
    $self->{block}->{path} = $blockfontpath;
  } else {
    croak "File $blockfontpath not found";
  }
  return $self->{block};
}

sub usefont {
  my ($self, $font) = @_;
  if (!defined $font || exists $self->{font}->{$font}) {
    $self->{fontname} = $font;
  } else {
    croak "Unknown font $font";
  }
  return $self->{fontname};
}

sub font {
  my ($self, $fontpath, $fontname, $info) = @_;
  if (-e $fontpath) {
    if (defined $fontname) {
      $self->{fontname} = $fontname;
    } else {
      croak 'Fontname not present';
    }
    if (!defined $self->{font}->{$fontname}) {
      $self->{font}->{$self->{fontname}} = setnewfont($self, $self->{fontname});
    }
    amendfont($self, $fontname, $info) if defined $info;
    $self->{font}->{$fontname}->{path} = $fontpath;
  } else {
    croak "File $fontpath not found"
  }
  return $fontpath, $fontname, $self->{font}->{$fontname};
}

sub amendfont {
  my ($self, $fontname, $info) = @_;
  if (exists $self->{font}->{$fontname}) {
    if (ref $info eq 'HASH') {
      my %fontkeys = map { $_ => 1} @{$self->{afontkeys}};

      while (my ($key, $value) = each %{$info}) {
        if (exists $fontkeys{$key}) {
          if ((!defined $value && $key =~ /^front$|^back$|^under$/) || isnumeric($value)) {
            $self->{font}->{$fontname}->{$key} = $value;
          } else {
            croak "Font setting for $key is not recognised as a decimal number";
          }
        } else {
          croak "Unknown font setting '$key'";
        }
      }

    } else {
      croak "Use a hash reference to change font settings";
    }
  } else {
    croak "Unknown font $fontname";
  }
  return $self->{font}->{$fontname};
}

sub setnewfont {
  my ($self, $font) = @_;
  for ($font) {
    if (/^gotype$/)  { return setfontconstants($self, 221, 220, 19, 14, 1, -5, undef); }
    if (/^almanac$/) { return setfontconstants($self, 65, 68, 25, 19, -2.5, -3.5, undef); } # 61505 61508 ?
#    if (/^almanac$/) { return setfontconstants($self, 61505, 61508, 25, 19, -2.5, -3.5, undef); } # try this if above fails # both work there is a duplicate set of characters one in the private use area
    if (/^gobd$/)    { return setfontconstants($self, 61491, 61490, 21, 16, 0, -3, undef); }
  }
  return setfontconstants($self, undef, undef, 1, 1, 0, 0, undef);
}

sub setfontconstants {
  my $self = shift;
  my %hash;
  @hash{@{$self->{afontkeys}}} = @_;
  return \%hash
}

sub extendboard {
  my ($self, $bflag) = @_;
  return set($self, 'bflag', $bflag);
}

sub checkselfcapture {
  my ($self, $flag) = @_;
  return set($self, 'checkselfcapture', $flag);
}

sub maxmoves {
  my ($self, $mm) = @_;
  if (defined $mm) {
    if ($mm !~ /^\d+$/) {
      croak "Maxmoves must be an integer greater than or equal to 0 (got '$mm')";
    }
  }
  return set($self, 'maxmoves', $mm);
}

sub scale {
  my ($self, $scale) = @_;
  if (defined $scale) {
    if (isnumeric($scale) && $scale > 0) {
      $self->{stonesize} *= $scale/128;
    } else {
      croak "Scale must be a number greater than 0 (got '$scale')";
    }
  }
  return set($self, 'scale', $scale);
}

sub isnumeric {
  my $string = shift;
  scalar $string =~ /^-?\d+(\.\d+)?$/;
}

sub colours {
  my ($self, $colours) = @_;
  if (defined $colours) {
    if (ref $colours eq 'HASH') {

      while (my ($key, $value) = each %{$colours}) {
        if (defined $value) {
          if ($value =~ /^[\da-f]{6}$/ || $value eq 'random') {
            if (exists $self->{colourhash}->{$key}) {
              $self->{colourhash}->{$key} = $value;
            } else {
              croak "Colour element '$key' not found";
            }
          } else {
           croak "Colour value '$value' not a colour";
          }
        } else {
          croak "no colour provided for $key";
        }
      }

    } else {
      croak 'not an anonymous hash';
    }
  }

  return $self->{colourhash}
}

sub set {
  my ($self, $index, $value) = @_;
  $self->{$index} = $value if defined $value;
  return $self->{$index};
}

sub makeimage {
  my ($self, $in, $out) = @_;
  croak 'sgf file not entered' if !defined $in;
  croak "File $in not found" if !-e $in;
  croak 'No output file specified' if !defined $out;
  my ($sizeref, $tsumego, $aref, $nomove) = process_sgf($in, $self->{maxmoves}, $self->{rflag}, $self->{checkselfcapture});
  my $st = $self->{stonesize};
  my @sizes = (($sizeref->[0] + 1) * $st, ($sizeref->[1] + 1) * $st);
  $self->{sizes} = \@sizes;
  my $code = $self->{code};
  $code->setsize(@sizes);
  setcolours($self, $tsumego, $nomove);
  drawboard($self, $st, @sizes);
  drawstones($self, $aref, $tsumego, $st);
  $code->frame($self, @sizes) if $self->{bflag};
  crop($self, $st, $aref, @sizes) if $tsumego || $nomove;
  $code->resize($self->{scale});
  $code->create($out);
  $code->destroy();
}

sub getrandom {
  join '', map { sprintf '%x', int(rand(16)) } (0..5);
}

sub setcolours {
  my ($self, $tsumego, $nomove) = @_;
 
  while (my ($colour, $value) = each %{$self->{colourhash}}) {
    if (defined $value) {
      $value = getrandom() if $value eq 'random';
      $self->{colours}->{$colour} = $self->{code}->setcolour($value);
    }
  }

  my $colourtype = $tsumego ? 'tsumego' : $nomove ? 'nomove' : 'game';
  $self->{colours}->{board} = $self->{colours}->{$colourtype};
  $self->{colours}->{offboard} = $self->{colours}->{"${colourtype}off"};
}

sub drawstones {
  my ($self, $aref, $tsumego, $st) = @_;
  my ($factor, $fontinfo);
  if (defined $self->{fontname}) {
    $fontinfo = $self->{font}->{$self->{fontname}};
    my $modfactor = $self->{module} eq 'GD' ? 'gdfactor' : 'imfactor';
    $factor = $fontinfo->{$modfactor} * $st/20;
  }
  my %mc = map {$_ => makecolours($_, $self->{colours})} ('b', 'w');

  for (@{$aref}) {
    my $prefix = getprefix(lc $_->[0], $tsumego);
    if (defined $prefix) {
      my @p = ($st, $self->{outline}, centre2($st, $_->[1], $_->[2]), @{$mc{$prefix}});
      if (defined $self->{fontname}) {
        if (defined $fontinfo->{black} && defined $fontinfo->{white}) {
          stonewithfontblackorwhite($self->{code}, @p, $fontinfo, $factor, $prefix, $self);          
        } else {
          stonewithfont($self->{code}, @p, $fontinfo, $factor);
        }
      } else {
        $self->{code}->stone($self->{stonefactor}, @p);
      }
    }
  }

}
 
sub makecolours {
  my ($prefix, $scol) = @_;
  my @colours;

  for ('so', 's', 'under') {
    push @colours, $scol->{"$prefix$_"};
  }

  return \@colours;
}

sub getprefix {
  my ($tag, $tsumego) = @_;
  if ($tag =~ /^(a?)(b|w)/) {
    unless ($tsumego && !defined $1) {
      return $2;
    }
  } else {
    croak "Unknown tag '$tag'";
  }
}

sub stonewithfontblackorwhite {
  my ($code, $st, $outline, $x, $y, $colour, $bgcolour, $uncolour, $finfo, $factor, $prefix, $self) = @_;
  my $blackorwhite = $prefix eq 'b' ? 'black' : 'white';
  if (defined $finfo->{$blackorwhite}) {
    if (defined $self->{block}) {
      if (defined $self->{block}->{character}) {
        my $blockx = $x + ($self->{block}->{xnudge} * $st/20);
        my $blocky = $y + ($self->{block}->{ynudge} * $st/20);
        $code->stonef($blockx, $blocky, $uncolour, $self->{block}->{factor} * $st/20, $st, $self->{block}->{character}, $self->{block}->{path});
      } else {
        $code->square($uncolour, $x + $st/2, $y + $st/2, $x - $st/2, $y - $st/2);
      }
    }
#    # code to rotate black symbols
#    if ($blackorwhite eq 'black') {
#      $code->rotate();
#      my $xtransform = $self->{sizes}->[0]/2;
#      my $ytransform = $self->{sizes}->[1]/2;
#      $x -= $xtransform;
#      $y -= $ytransform;
#      ($x, $y) = (-$x, -$y);
#      $x += $xtransform;
#      $y += $ytransform;
#    }
    $x += $finfo->{xnudge} * $st/20;
    $y += $finfo->{ynudge} * $st/20;
    $code->stonef($x, $y, $colour, $factor, $st, $finfo->{$blackorwhite}, $finfo->{path});
#    if ($blackorwhite eq 'black') {
#      $code->rotate();
#    }
  }
}

sub stonewithfont {
  my ($code, $st, $outline, $x, $y, $fgcolour, $bgcolour, $uncolour, $finfo, $factor) = @_;
  $x += $finfo->{xnudge} * $st/20;
  $y += $finfo->{ynudge} * $st/20;
  if (defined $finfo->{under}) {
    $code->stonef($x, $y, $uncolour, $factor, $st, $finfo->{under}, $finfo->{path});
  }
  if (defined $finfo->{back}) {
    $code->stonef($x, $y, $bgcolour, $factor, $st, $finfo->{back}, $finfo->{path});
  }
  if (defined $finfo->{front} && $outline) {
    $code->stonef($x, $y, $fgcolour, $factor, $st, $finfo->{front}, $finfo->{path});
  }
}

sub drawboard {
  my ($self, $st, $x, $y) = @_;
  my $bo = $st/2;
  my @p = ($bo, $bo, $x - $bo, $y - $bo);
  my $code = $self->{code};
  $code->board($self, $x, $y, $bo, @p);
  my $linecolour = $self->{colours}->{line};
  drawdots($code, $x/$st, $y/$st, $linecolour, $st);
  drawlines($code, $self->{module}, $x/$st, 'v', $y - $st, $linecolour, $st); # vertical
  drawlines($code, $self->{module}, $y/$st, 'h', $x - $st, $linecolour, $st); # horizontal
}

sub drawdots {
  my ($code, $x, $y, $linecolour, $st) = @_;
  my @dots = getdots($x, $x == $y);
  if (@dots) {

    for my $dot (@dots) {
      $code->dots($st, centre2($st, $dot->[0], $dot->[1]), $linecolour);
    }

  }
}

sub centre2 {
  return centre($_[0], $_[1]), centre($_[0], $_[2])
}

sub centre {
  ($_[1] + 0.5) * $_[0];
}

sub getdots {
  my ($x, $flag) = @_;
  if ($flag) {
    if ($x == 19) {
      return [3, 3], [3, 9], [3, 15], [9, 3], [9, 9], [9, 15], [15, 3], [15, 9], [15, 15];
    } elsif ($x == 13) {
      return [3, 3], [3, 9], [9, 3], [9, 9], [6, 6];
    } elsif ($x == 9) {
      return [4, 4];
    }
  } 
  return;
}

sub drawlines {
  my ($code, $module, $z, $way, $max, $colour, $st) = @_;
  # extend lines to make corners sharp
#  my ($soffset, $eoffset) = $module eq 'GD' ? (-2, 1) : (-2, 2);
  # that didn't work 
  my ($soffset, $eoffset) = (0, 0);
  # make the edge double thickness or it disappears when viewed as a thumbnail
  drawline($code, $way, $st/2, 0, 1, $max, $colour, $soffset, $eoffset);
  drawline($code, $way, $st/2, ($z - 1) * $st, -1, $max, $colour, $soffset, $eoffset);

  for (0..$z-1) {
    drawline($code, $way, $st/2, $_ * $st, 0, $max, $colour, $soffset, $eoffset);
  }

}

sub drawline {
  my ($code, $way, $boardoffset, $step, $nudge, $max, $colour, $soffset, $eoffset) = @_;
  my $snb = $step + $nudge + $boardoffset;
  my @p = ($snb, $boardoffset + $soffset, $snb, $boardoffset + $max + $eoffset);
  if ($way eq 'h') {
    push @p, shift @p;
  }
  $code->line($colour, $boardoffset, @p);
}

sub crop {
  my ($self, $st, $aref, $width, $height) = @_;
  my ($minx, $maxx, $maxxflag) = findminmax($width/$st - 1, $aref, 1);
  my ($miny, $maxy, $maxyflag) = findminmax($height/$st - 1, $aref, 2);
  my ($geowidth, $geoheight) = ($st * ($maxx - $minx + 1), $st * ($maxy - $miny + 1));
  my $xoffset = my $yoffset = 0;
  if ($self->{bflag}) {
    my $stext = $st * $self->{extension};
    $geowidth += $stext if !$minx;
    $geoheight += $stext if !$miny;
    if ($maxxflag) {
      $geowidth += $stext;
      $xoffset = $stext if $minx;
    }
    if ($maxyflag) {
      $geoheight += $stext;
      $yoffset = $stext if $miny;
    }
  }
  if ($geowidth != $width || $geoheight != $height) { 
    $self->{code}->crop($geowidth, $geoheight, $xoffset + $minx * $st, $yoffset + $miny * $st);
  }
}

sub findminmax {
  my ($maxpossible, $aref, $pos) = @_;
  my @a = map { $_->[$pos] } @{$aref};
  return findmin(@a), findmax($maxpossible, @a);
}

sub findmin { # expand to the edge if nearby
  my $min = min @_;
  $min <= 2 ? 0 : $min - 1;
}

sub findmax { # expand to the edge if nearby
  my $maxpossible = shift;
  my $max = max @_;
  $max >= $maxpossible - 2 ? ($maxpossible, 1) : ($max + 1, 0);
}

1;

__END__

=head1 NAME

    sgfThumbnailer - create images of sgf files suitable for use as thumbnails

=head1 SYNOPSIS

    use sgfThumbnailer;
    my $writer = new sgfThumbnailer();
    $writer->makeimage('in.sgf', 'out.png');

=head1 DESCRIPTION

    sgfThumbnailer creates a quick png image from an sgf file.
    Included is a script to set up automatic thumbnail generation for linux users within the Nautilus file manager.
    Move numbers are not supported, neither are variations.
    You can choose which graphics library to use (GD or Image::Magick), whichever is available. 
    Games that are classed as tsumego (ie problems) will not display moves, only initial setup.

=head1 METHODS

=head3 new

    my $writer = new sgfThumbnailer();
    Calling new with no parameter causes sgfThumbnailer to try to load GD, and then Image::Magick if that fails.

    my $writer = new sgfThumbnailer($lib);
    Try to load graphics library $lib ('GD' or 'Image::Magick')

=head1 OPTIONS

=head3 setlib

    Set the graphics library to either 'Image::Magick' or 'GD'.

    $writer->setlib('GD');

    GD is faster, but Image::Magick produces better looking images.
    If you have suitable fonts (see below), GD can produce thumbnails of the same quality.

    Returns the library in use.

=head3 font

    Use a font to draw the stones. This is useful if your version of GD can't antialias circles, but can antialias text.
    To use one of three preconfigured fonts (Alamanc MT, GoType New and Gobd), use font(path, name), where name is 'almanac', 'gotype' or 'gobd'.

    If you want to use a different font, use font(path, name, hashref). Name is any identifying string.
    The following hash reference keys are used

    front (the outline character code, which may look like a white stone)
    back (the fill character code, which may look like a black stone)
    imfactor (Image::Magick scale factor)
    gdfactor (GD scale factor)
    xnudge (+ve numbers nudge right)
    ynudge (+ve numbers nudge down)
    under (second background character code, which can be used to make three coloured stones)
    black (instead of front, back and under, this key can be used to set the character to represent a black move)
    white (to represent a white move. black and white will be coloured according to bunder and wunder)
    
    eg $writer->font('/path/to/wingding.ttf', 'wd', {front => 123, back => 124, imfactor => 27, gdfactor => 20, xnudge => -1, ynudge => -1});

    Characters are drawn in the order under, back, front.
    The default settings for a non-preconfigured font are -
    {front => undef, back => undef, under => undef, imfactor => 1, gdfactor => 1, xnudge => 0, ynudge => 0}

    Once black and white have been defined, they will be used in preference to front, back, under until undefined.

    Returns an array of path, name, hash reference.

=head3 amendfont

    $writer->amendfont($fontname, $hashref);

    Amend the the details of the font $fontname according to the contents of $hashref;

    eg $writer->amendfont('wd', {front => 122});

    Returns a hash reference of the current font settings.

=head3 usefont

    $writer->usefont($fontname);

    Declare that you want to draw images using $fontname. If $fontname is undefined, fonts will not be used to draw the image.
    If usefont is not called, the font mentioned in the last call to font() or amendfont() will be used.

    eg $writer->usefont('gotype');

    Returns the name of the font being used.

=head3 block

    $writer->block($fontpath, {character => $character, factor => 44, xnudge => $xnudge, ynudge => $ynudge});

    Use a different font and character to draw a background $character when using the black and white font keys.
    Coloured according to bunder and wunder - see colours heading.
    When character is undefined, a square will be drawn.

    eg $writer->block('/usr/share/fonts/truetype/msttcorefonts/Arial.ttf', {character => 0x25a0, factor => 44, xnudge => -3, ynudge => 0});

    ('0x25a0' represents a square that can be coloured)

    Returns a hashref of the current block settings.

=head3 colours

    $writer->colours({element => colour});

    Associates a board element with a colour. The element codes used are

    bs (black stone)
    bso (black stone outline)
    ws (white stone)
    wso (white stone outline)
    line (board lines)
    tsumego (tsumego games)
    tsumegooff (the outer edge of the image)
    nomove (tsumego with no answer!)
    nomoveoff (the outer edge of the image)
    game (ordinary games)
    gameoff (the outer edge of the image)
    bunder (black under character)
    wunder (white under character)
    
    Default colors are as though
    $writer->colours({tsumego => 'cfd5e5', nomove => 'ff739b', game => 'c9c5a4', line => '000000', bs => '000000', ws => 'ffffff', bso => '000000', wso => '000000', bunder => 'ffffff', wunder => 'ffffff', tsumegooff => 'c7f4f3', nomoveoff => 'ffffff', gameoff => 'f39d2a', offboard => 'ffffff'});

    Setting the colour to 'random' will cause a random colour to be used.
    eg $writer->colours({tsumego => 'ffff00', bs => 'random'})

    Returns a hash reference of the colours set

=head3 extendboard

    $writer->extendboard($value);

    If $value is true, the image be extended to resemble a real board.
    If its false (default), the image will not extend beyond a stone on the edge.
    Returns the current value.

=head3 extension

    $writer->extension($value);

    Sets how far the board extends past a stone on the edge, in fractions of a stone.
    If extendboard is false, this value has no effect.
    The default is 0.12
    Returns the current value.

=head3 maxmoves

    $writer->maxmoves($integer);

    Set the maximum number of moves (not including setup) to illustrate. Default is 25.
    Returns the current value.

=head3 stonesize

    $writer->stonesize($number);

    Set the relative size of stones. Default is 0.92, to create a small gap between the stones.
    Has no effect when using a font (use gdfactor and imfactor instead.)
    Returns the current value.

=head3 outline

    $writer->outline($value);

    If $value is true (default), stones will be drawn with an outline.
    If its false, stones will not be drawn with an outline.
    If using fonts, you could also use $writer->font('/path/to/somefont.ttf', 'myfont', {front => undef});
    Returns the current value.

=head3 realboard

    $writer->realboard($value);

    If $value is true, captured stones will not be drawn.
    If its false (default), captured stones will be drawn.
    Returns the current value.

=head3 scale

    $writer->scale($integer);

    Set or get scale. Default is 128.
    Images will have a maximum height and width of $value.
    Returns the current value.

=head3 checkselfcapture

    $writer->checkselfcapture($value);

    If $value is true, and realboard is true, a check will be made to see if a self capture has happened.
    If its false (default), no such check will be made, (which is a little faster.)
    Returns the current value.

=cut
